from django.db import models
from mptt.models import MPTTModel, TreeForeignKey
# from users.models import User
from django.db.models.signals import post_save

from projects.models import ProjectGroup
from users.models import User


class Company(MPTTModel):
    name = models.CharField(
        'наименование',
        max_length=255,
        unique=True,
        db_index=True,
    )
    full_name = models.TextField(
        'полное наименование',
        unique=True
    )
    postal_address = models.TextField('почтовый адрес', blank=True)
    legal_address = models.TextField('юридический адрес', blank=True)
    parent_company = TreeForeignKey(
        'self',
        on_delete=models.CASCADE,
        verbose_name='головная компания',
        related_name='subsidiaries',
        blank=True,
        null=True
    )
    locations = models.ManyToManyField(
        'core.Location',
        related_name='+',
        blank=True,
        verbose_name='местоположения'
    )

    class Meta:
        verbose_name_plural = 'Companies'

    class MPTTMeta:
        parent_attr = 'parent_company'

    def __str__(self):
        return self.name

    @classmethod
    def post_create(cls, sender, instance, created, *args, **kwargs):
        """
        Method for auto-creating default project group for company, triggered by django signals.
        """
        if not created:
            return
        ProjectGroup.objects.create(
            name=f'{instance.name} - проекты',
            company=instance,
        )


post_save.connect(Company.post_create, sender=Company)


class Department(MPTTModel):
    name = models.CharField('наименование', max_length=255)
    company = models.ForeignKey(
        'Company',
        on_delete=models.CASCADE,
        related_name='departments',
        verbose_name='компания'
    )
    parent_department = TreeForeignKey(
        'self',
        on_delete=models.CASCADE,
        related_name='subdepartments',
        verbose_name='головной департамент',
        blank=True,
        null=True
    )

    class Meta:
        unique_together = ('name', 'company', 'parent_department')

    class MPTTMeta:
        parent_attr = 'parent_department'

    def __str__(self):
        return self.name

    def get_managers(self):
        """
        Gets the list of this department managers
        :return: list of managers
        :rtype: list of User (Queryset)
        """
        return self.employees.filter(department_manager=True)

    # def get_all_managers(self, include_self=True):
    #     """
    #     Gets the list of this and all children departments managers
    #
    #     :param include_self: include managers from this department in output
    #     :type include_self: bool
    #     :return: list of managers
    #     :rtype: list of User (Queryset)
    #     """
    #     return User.objects.filter(
    #         department__in=self.get_descendants(include_self=include_self),
    #         department_manager=True
    #     )

    def get_all_users(self):
        """
        Gets all users of this department and all children departments
        :return: queryset list of users
        """
        return User.objects.filter(
                    department__in=self.get_descendants(include_self=True),
                )


class Position(models.Model):
    title = models.CharField(
        'должность',
        unique=True,
        max_length=255,
    )

    def __str__(self):
        return self.title


class CompanyCurrency(models.Model):
    company = models.ForeignKey(Company, on_delete=models.CASCADE, related_name='currencies')
    currency = models.CharField('ISO-код валюты', default='RUB', max_length=3)
    main = models.BooleanField('основная валюта', null=False, blank=False, default=False)

    class Meta:
        unique_together = ('company', 'currency')

    def __str__(self):
        return f'{self.company.name} - {self.currency}'

