from django.test import TestCase
from users.models import User
from .models import Department, Company
from rest_framework.test import APIRequestFactory, force_authenticate, APIClient
from django.conf import settings


class DepartmentTestCase(TestCase):
    def setUp(self):
        self.company = Company.objects.create(name='comp', full_name='comp')
        self.departments = {
            'dep1': Department.objects.create(name='dep1', company=self.company),
            'dep2': Department.objects.create(name='dep2', company=self.company),
            'dep3': Department.objects.create(name='dep3', company=self.company),
            'dep4': Department.objects.create(name='dep4', company=self.company),
        }
        self.departments['dep3'].parent_department = self.departments['dep1']
        self.departments['dep3'].save()
        self.departments['dep4'].parent_department = self.departments['dep3']
        self.departments['dep4'].save()

        self.users = {
            'dep1_usr1': User.objects.create(email='dep1_usr1', department=self.departments['dep1']),
            'dep1_usr2': User.objects.create(email='dep1_usr2', department=self.departments['dep1']),
            'dep1_usr3': User.objects.create(email='dep1_usr3', department=self.departments['dep1']),
            'dep1_head1': User.objects.create(email='dep1_head1', department=self.departments['dep1'],
                                              department_manager=True),
            'dep1_head2': User.objects.create(email='dep1_head2', department=self.departments['dep1'],
                                              department_manager=True),
            'dep2_usr1': User.objects.create(email='dep2_usr1', department=self.departments['dep2']),
            'dep2_head1': User.objects.create(email='dep2_head1', department=self.departments['dep2'],
                                              department_manager=True),
            'dep3_usr1':  User.objects.create(email='dep3_usr1', department=self.departments['dep3']),
            'dep3_head1': User.objects.create(email='dep3_head1', department=self.departments['dep3'],
                                              department_manager=True),
            'dep4_usr1':  User.objects.create(email='dep4_usr1', department=self.departments['dep4']),
            'dep4_head1': User.objects.create(email='dep4_head1', department=self.departments['dep4'],
                                              department_manager=True),
            'dep4_head2': User.objects.create(email='dep4_head2', department=self.departments['dep4'],
                                              department_manager=True),
        }

    def test_get_managers(self):
        self.assertQuerysetEqual(
            self.departments['dep1'].get_managers(),
            [
                self.users['dep1_head1'],
                self.users['dep1_head2'],
            ],
            ordered=False,
            transform=lambda x: x
        )
        self.assertQuerysetEqual(
            self.departments['dep2'].get_managers(),
            [
                self.users['dep2_head1'],
            ],
            ordered=False,
            transform=lambda x: x,
        )

    def test_get_all_managers(self):
        self.assertQuerysetEqual(
            self.departments['dep1'].get_all_managers(),
            [
                self.users['dep1_head1'],
                self.users['dep1_head2'],
                self.users['dep3_head1'],
                self.users['dep4_head1'],
                self.users['dep4_head2'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.departments['dep3'].get_all_managers(),
            [
                self.users['dep3_head1'],
                self.users['dep4_head1'],
                self.users['dep4_head2'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.departments['dep4'].get_all_managers(),
            [
                self.users['dep4_head1'],
                self.users['dep4_head2'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.departments['dep1'].get_all_managers(include_self=False),
            [
                self.users['dep3_head1'],
                self.users['dep4_head1'],
                self.users['dep4_head2'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.departments['dep4'].get_all_managers(include_self=False),
            [],
            ordered=False,
            transform=lambda x: x,
        )


class CompanyTestCase(TestCase):
    def setUp(self):
        self.company = Company.objects.create(name='comp', full_name='comp')
        self.user = User.objects.create(email='user1')

    def test_auto_create_project_group(self):
        pg = self.company.project_groups.all()
        self.assertEqual(pg.count(), 1)
        self.assertEqual(pg[0].name, f'{self.company.name} - проекты')

    def test_auto_create_user_role(self):
        client = APIClient()
        client.force_authenticate(user=self.user)
        client.post('/companies/', {'name': 'comp1', 'full_name': 'comp1'})
        role_id = settings.DEFAULT_ROLES['COMPANY_ADMIN']
        comp1 = Company.objects.get(name='comp1')
        self.assertEqual(comp1.users_roles.count(), 1)
        self.assertEqual(comp1.users_roles.all()[0].user, self.user)
        self.assertEqual(comp1.users_roles.all()[0].role.id, role_id)
