from django.contrib.contenttypes.models import ContentType
from drf_yasg.utils import swagger_auto_schema
from drf_yasg.openapi import Parameter
from collections import defaultdict

from easyaudit.models import CRUDEvent
from rest_framework.decorators import action
from rest_framework.viewsets import ModelViewSet


class RelationFieldsSerializerMixin(object):

    def __init__(self, *args, **kwargs):
        include = kwargs.pop('include', [])

        includes = defaultdict(list)
        for inc in include:
            fs = inc.split('__', 1)
            if len(fs) == 1:
                if fs[0] not in includes:
                    includes[fs[0]] = []
            else:
                includes[fs[0]].append(fs[1])
        super(RelationFieldsSerializerMixin, self).__init__(*args, **kwargs)

        for field in self.get_relation_fields():
            if field[0] in self.fields:
                model = getattr(self.Meta, "model")
                if model.__dict__[field[0]].__class__.__name__ in ['ReverseManyToOneDescriptor', 'ManyToManyDescriptor']:
                    many = True
                else:
                    many = False
                if field[0] in includes:
                    if RelationFieldsSerializerMixin in field[1].__mro__:
                        self.fields['rel_' + field[0]] = field[1](
                            read_only=True,
                            source=field[0],
                            include=includes[field[0]],
                            required=False,
                            many=many)
                    else:
                        self.fields['rel_' + field[0]] = field[1](read_only=True, required=False, source=field[0], many=many)
                else:
                    self.fields['rel_' + field[0]] = field[1](read_only=True, many=many, required=False)

    def get_relation_fields(self):
        if hasattr(self, "Meta"):
            return getattr(self.Meta, "relation_fields", set())
        return set()


class RelationFieldsViewMixin(object):

    def get_serializer(self, *args, **kwargs):

        serializer_class = self.get_serializer_class()
        if not serializer_class:
            return None
        include = []
        if self.request.method == 'GET':
            include = self.request.query_params.get("include", [])

            if include:
                include = include.replace(' ', '').split(',')

        kwargs['context'] = self.get_serializer_context()
        if include:
            kwargs['include'] = include
        return serializer_class(*args, **kwargs)

    @swagger_auto_schema(manual_parameters=[Parameter(
        'include',
        'query',
        description='Relation fields to include in response',
        required=False,
        type='string')])
    def retrieve(self, *args, **kwargs):
        return super(RelationFieldsViewMixin, self).retrieve(*args, **kwargs)

    @swagger_auto_schema(manual_parameters=[Parameter(
        'include',
        'query',
        description='Relation fields to include in response',
        required=False,
        type='string')])
    def list(self, *args, **kwargs):
        return super(RelationFieldsViewMixin, self).list(*args, **kwargs)


class MultiSerializerViewSetMixin(object):
    serializers = {
        'default': None,
    }

    def get_serializer_class(self):
        return self.serializers.get(self.action, self.serializers['default'])


# class HistoryViewViewSetMixin(ModelViewSet):
#     @swagger_auto_schema(
#         operation_description='Object history',
#         responses={200: HistorySerializer(many=True)})
#     @action(
#         detail=True,
#         methods=['GET'],
#     )
#     def history(self, request, pk):
#         model = getattr(self.get_serializer_class().Meta, 'model')
#
#         CRUDEvent.objects.filter(
#             content_type=ContentType.objects.get_for_model(model),
#             object_id=str(pk),
#         )
