from typing import Optional

from django.db import models
from mptt.models import MPTTModel, TreeForeignKey
from django.utils import timezone
from softdelete import models as sd_models
from django.apps import apps
from uuid import uuid4

import projects.models
from core import helpers


class DocumentType(MPTTModel):
    name = models.CharField('наименование', max_length=255)
    order = models.PositiveSmallIntegerField('порядок', default=20)
    parent_type = TreeForeignKey(
        'self',
        verbose_name='родительский тип',
        related_name='subtypes',
        on_delete=models.PROTECT,
        blank=True,
        null=True,
    )
    meta_fields = models.ManyToManyField(
        'DocumentMetaField',
        related_name='document_types',
        blank=True
    )

    class MPTTMeta:
        parent_attr = 'parent_type'

    def __str__(self):
        return self.name


class Document(sd_models.SoftDeleteObject, models.Model):
    created = models.DateTimeField(default=timezone.now)
    author = models.ForeignKey('users.User', null=False, blank=False, related_name='documents',
                               on_delete=models.PROTECT)
    name = models.CharField('наименование', max_length=255)
    description = models.TextField('описание', blank=True)
    document_type = models.ForeignKey(DocumentType, verbose_name='тип документа', null=False, blank=False,
                                      on_delete=models.PROTECT)

    number = models.CharField('номер документа', blank=True, max_length=255)
    date = models.DateField('дата документа', null=True, blank=True)

    projects = models.ManyToManyField('projects.Project', related_name='documents')
    project_stages = models.ManyToManyField('projects.ProjectStage', related_name='documents', blank=True)

    # TODO: add works
    # works = models.ManyToManyField('')

    def __str__(self):
        return self.name

    # TODO: implement
    def get_name(self):
        pass

    def get_last_version_file(self):
        return self.files.last()

    def get_last_version_number(self):
        doc_file: DocumentFile = self.get_last_version_file()
        if doc_file:
            return doc_file.version
        else:
            return 0


class DocumentFile(sd_models.SoftDeleteObject, models.Model):
    document = models.ForeignKey(Document, on_delete=models.CASCADE, related_name='files', verbose_name='файл', blank=True)
    version = models.PositiveSmallIntegerField('версия')
    author = models.ForeignKey('users.User', related_name='document_author', verbose_name='автор', on_delete=models.PROTECT)
    created = models.DateTimeField(default=timezone.now)

    doc_file = models.FileField(
        upload_to=helpers.get_file_path,
        null=True,
        blank=True
    )

    MIME = models.CharField('MIME', max_length=255, blank=True)
    extension = models.CharField('расширение', max_length=10)
    comment = models.TextField('описание файла', blank=True)
    original_name = models.TextField('название при загрузке', blank=True)

    def __str__(self):
        return '{} - в{}'.format(self.document, self.version)

    class Meta:
        unique_together = ('version', 'document')
        ordering = ('document', 'version')


class DocumentMetaField(models.Model):
    DATE = 'DT'
    CHAR = 'CH'
    INT = 'INT'
    FLT = 'FLT'
    TXT = 'TXT'
    FK = 'FK'
    MTM = 'MTM'

    META_TYPE_CHOICES = (
        (DATE, 'Дата'),
        (CHAR, 'Короткий текст'),
        (TXT, 'Длинный текст'),
        (INT, 'Целое число'),
        (FLT, 'Дробное число'),
        (FK, 'Отношение'),
        (MTM, 'Множественное отношение'),
    )

    name = models.CharField(
        'имя метаданных',
        max_length=50,
        db_index=True
    )

    meta_type = models.CharField(
        'тип метаданных',
        max_length=20,
        db_index=True,
        choices=META_TYPE_CHOICES
    )
    in_name = models.BooleanField('отображать в имени файла', default=True)
    in_name_prefix = models.CharField('префикс в для имени', max_length=10, blank=True)
    required = models.BooleanField('обязательное поле', default=True)
    placeholder = models.CharField('подсказка', max_length=255, blank=True)
    order = models.IntegerField('порядок', default=10)

    model_name = models.CharField('имя модели', max_length=50, null=True, blank=True)
    model_app = models.CharField('имя приложения', max_length=50, null=True, blank=True)

    class Meta:
        ordering = ['order']

    def __str__(self):
        return self.name

    @property
    def model(self) -> Optional[models.Model]:
        if self.meta_type in ('FK', 'MTM'):
            try:
                return apps.get_model(app_label=self.model_app, model_name=self.model_name)
            except LookupError:
                return None
        return None


class DocumentMetaValue(models.Model):
    meta = models.ForeignKey(DocumentMetaField, on_delete=models.PROTECT, related_name='values')
    document = models.ForeignKey(Document, on_delete=models.CASCADE, related_name='metas')
    content = models.TextField('значение')
    object_id = models.BigIntegerField('идентификатор объекта', null=True, blank=True)

    def __str__(self):
        return f'{self.document} - {self.meta}'

    @property
    def display_value(self) -> str:
        if self.is_object:
            try:
                return str(self.object_model.objects.get(id=self.object_id))
            except (models.Model.DoesNotExist, AttributeError) as _:
                return ''
        return self.content

    @property
    def object_model(self) -> Optional[models.Model]:
        return self.meta.model_app

    @property
    def is_object(self) -> bool:
        return bool(self.object_id)







