import datetime

from django.db import models
from django.db.models import Max, Sum
from djmoney.models.fields import MoneyField
from moneyed import Money
import traceback


class BudgetCategory(models.Model):
    name = models.CharField('наименование категории', max_length=255)
    order = models.PositiveSmallIntegerField('порядок', default=10, blank=True)
    company = models.ForeignKey('companies.Company', on_delete=models.CASCADE, related_name='budget_categories')

    def __str__(self):
        return f'{self.company} - {self.name}'


class BudgetType(models.Model):
    name = models.CharField('наименование категории', max_length=255)
    order = models.PositiveSmallIntegerField('порядок', default=10, blank=True)
    category = models.ForeignKey(BudgetCategory, on_delete=models.PROTECT)
    code = models.CharField('шифр затрат', max_length=255, blank=True, null=True)

    def __str__(self):
        return f'{self.category} - {self.name}'

    class Meta:
        ordering = ('category', 'order')


class ProjectBudgetType(models.Model):
    budget_type = models.ForeignKey(BudgetType, on_delete=models.PROTECT, related_name='project_budget_types')
    code = models.CharField('шифр затрат', max_length=255, blank=True, null=True)
    project = models.ForeignKey('projects.Project', on_delete=models.CASCADE, related_name='project_budget_types')
    total_planned = MoneyField('бюджет статьи', max_digits=19, decimal_places=4, default=0)
    # class Meta:
    #     unique_together = ('budget_type', 'project', 'code')

    def __str__(self):
        return f'{self.project} - {self.budget_type}'

    @property
    def name(self):
        return self.budget_type.name

    @property
    def stats(self):
        res = {
            'plan': 0,
            'fact': 0,
            'forecast': 0,
            'current_fact_and_fc': 0,
            'fact_percent': 0,
            'plan_percent': 0,
            'plan_fact_remaining': 0,
            'plan_fact_remaining_percent': 0,
            'current_fact_and_fc_overplan': 0,
            'current_fact_and_fc_overplan_percent': 0,
            'total_overplan': 0,
            'total_overplan_percent': 0,
        }
        for i in self.items.all():
            res['plan'] += i.plan.amount

            res['forecast'] += i.forecast

            if i.approved:
                res['current_fact_and_fc'] += i.fact
                res['fact'] += i.fact
            else:
                res['current_fact_and_fc'] += i.forecast.amount

        res['fact_percent'] = res['fact'] / self.total_planned.amount
        res['plan_percent'] = res['plan'] / self.total_planned.amount

        res['plan_fact_remaining'] = self.total_planned.amount - res['fact']
        if res['plan_fact_remaining'] < 0:
            res['plan_fact_remaining'] = 0
        res['plan_fact_remaining_percent'] = res['plan_fact_remaining'] / self.total_planned.amount
        res['current_fact_and_fc_overplan'] = res['current_fact_and_fc'] - self.total_planned.amount
        res['current_fact_and_fc_overplan_percent'] = abs(res['current_fact_and_fc_overplan']) / self.total_planned.amount

        res['total_overplan'] = res['plan'] - self.total_planned.amount
        res['total_overplan_percent'] = abs(res['total_overplan']) / self.total_planned.amount

        return res


class ProjectBudgetItem(models.Model):
    date = models.DateField('дата')
    project = models.ForeignKey('projects.Project', on_delete=models.CASCADE, related_name='project_budget_items')
    budget_type = models.ForeignKey(ProjectBudgetType, on_delete=models.PROTECT, related_name='items')
    plan = MoneyField('план затрат', max_digits=19, decimal_places=4, default=0)
    forecast = MoneyField('прогноз затрат', max_digits=19, decimal_places=4, default=0)
    approved = models.BooleanField('утверждено', default=False)
    approved_by = models.ForeignKey(
        'users.User',
        verbose_name='кем утверждено',
        related_name='+',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    approved_at = models.DateField('утверждено в', blank=True, null=True)

    class Meta:
        unique_together = ('date', 'project', 'budget_type')
        ordering = ('date', )
        permissions = [
            ('approve_projectbudgetitem', 'Can approve project budget'),
            ('unapprove_projectbudgetitem', 'Can unapprove project budget'),
        ]

    def __str__(self):
        return f'{self.date} - {self.project}'

    @property
    def fact(self):
        return self.expenses.all().aggregate(Sum('value'))['value__sum'] or 0


class ProjectBudget(models.Model):
    project = models.OneToOneField(
        'projects.Project',
        on_delete=models.CASCADE,
        related_name='project_budget',
    )
    code = models.CharField('шифр затрат', max_length=255, blank=True)
    period_start = models.DateField('начало бюджетного периода', blank=True, null=True)
    period_end = models.DateField('конец бюджетного периода', blank=True, null=True)
    total_planned = MoneyField('бюджет проекта', max_digits=19, decimal_places=4, default=0)
    approved = models.BooleanField('утверждено', default=False)
    approved_by = models.ForeignKey(
        'users.User',
        verbose_name='кем утверждено',
        related_name='+',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    approved_at = models.DateField('утверждено в', blank=True, null=True)

    class Meta:
        permissions = [
            ('approve_projectbudget', 'Can approve project budget'),
            ('unapprove_projectbudget', 'Can unapprove project budget'),

        ]

    def __str__(self):
        return f'Бюджет проекта {self.project}'

    @property
    def stats(self):
        res = {
            'plan': 0,
            'fact': 0,
            'fact_percent': 0,
            'forecast': 0,
            'current_fact_and_fc': 0,
            'discipline': 0,
        }
        plan_approved = 0
        for i in self.project.project_budget_items.all():
            res['plan'] += i.plan.amount
            res['forecast'] += i.forecast.amount

            if i.approved:
                res['fact'] += i.fact
                res['current_fact_and_fc'] += i.fact
                plan_approved += i.plan.amount
            else:
                res['current_fact_and_fc'] += i.forecast.amount

            # now = datetime.datetime.now().replace(day=1).date()
            # if i.date < now:
            #     res['current_fact_and_fc'] += i.fact
            # else:
            #     res['current_fact_and_fc'] += i.forecast.amount

            res['fact_percent'] = 100 * res['fact'] / self.total_planned.amount
            res['discipline'] = 100 * res['fact'] / (plan_approved or 1)

        return res


class ProjectExpense(models.Model):
    budget_item = models.ForeignKey(
        ProjectBudgetItem,
        on_delete=models.CASCADE,
        related_name='expenses',
    )
    value = MoneyField('трата', max_digits=19, decimal_places=4, default=0)
    description = models.TextField('описание', blank=True, null=True)
    date = models.DateField('дата')

    class Meta:
        ordering = ('-date', )

    @property
    def budget_type_name(self):
        return self.budget_item.budget_type.name

    @property
    def budget_type(self):
        return self.budget_item.budget_type.id

    @property
    def approved(self):
        return self.budget_item.approved

    @property
    def editable(self):
        return not self.budget_item.approved

    @property
    def period(self):
        return self.date.replace(day=1)
