from typing import Union

from django.db.models import QuerySet
from drf_yasg.utils import swagger_auto_schema
from rest_framework import serializers
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.decorators import action
from django.utils import timezone

from core.mixins import MultiSerializerViewSetMixin
from core.viewsets import ModelViewSet
from .filters import ProjectExpenseFilter, ProjectBudgetItemFilter
from .models import BudgetCategory, BudgetType, ProjectBudgetType, ProjectBudgetItem, ProjectBudget, ProjectExpense
from .serializers import \
    BudgetCategorySerializer, BudgetTypeSerializer, ProjectBudgetTypeSerializer, ProjectBudgetItemSerializer, \
    ProjectBudgetSerializer, ProjectExpenseSerializer


class BudgetCategoryViewSet(ModelViewSet):
    serializer_class = BudgetCategorySerializer

    def get_queryset(self) -> QuerySet:
        return BudgetCategory.objects.filter(company=self.request.user.company)


class BudgetTypeViewSet(ModelViewSet):
    serializer_class = BudgetTypeSerializer

    def get_queryset(self) -> QuerySet:
        return BudgetType.objects.filter(category__company=self.request.user.company)


class ProjectBudgetTypeViewSet(ModelViewSet):
    serializer_class = ProjectBudgetTypeSerializer
    filterset_fields = ('project', )

    def get_queryset(self) -> QuerySet:
        return ProjectBudgetType.objects.filter(budget_type__category__company=self.request.user.company)


class ProjectBudgetItemViewSet(MultiSerializerViewSetMixin, ModelViewSet):
    filterset_class = ProjectBudgetItemFilter
    serializers = {'default': ProjectBudgetItemSerializer, 'toggle_approval': None}

    def get_queryset(self) -> QuerySet:
        return ProjectBudgetItem.objects.filter(budget_type__budget_type__category__company=self.request.user.company)

    @swagger_auto_schema(operation_description='Approve project budget item')
    @action(
        detail=True,
        methods=['POST'],
    )
    def toggle_approval(self, request, pk):
        pb = self.get_object()
        if pb.approved:
            pb.approved = False
            pb.approved_at = None
            pb.approved_by = None
        else:
            pb.approved = True
            pb.approved_at = timezone.now()
            pb.approved_by = request.user
        pb.save()
        return Response()


class ProjectBudgetViewSet(MultiSerializerViewSetMixin, ModelViewSet):
    filterset_fields = ('project', )
    serializers = {'default': ProjectBudgetSerializer, 'toggle_approval': None}

    def get_queryset(self) -> QuerySet:
        return ProjectBudget.objects.filter(project__project_group__company=self.request.user.company)

    @swagger_auto_schema(
        operation_description='Approve project budget',
        responses={200: ProjectBudgetSerializer(many=False)})
    @action(
        detail=True,
        methods=['POST'],
    )
    def toggle_approval(self, request, pk):
        pb = self.get_object()
        if pb.approved:
            pb.approved = False
            pb.approved_at = None
            pb.approved_by = None
        else:
            pb.approved = True
            pb.approved_at = timezone.now().date()
            pb.approved_by = request.user
        pb.save()
        return Response(ProjectBudgetSerializer(pb).data)


class ProjectExpenseViewSet(ModelViewSet):
    serializer_class = ProjectExpenseSerializer
    filterset_fields = ('budget_item', 'budget_item__project')

    def get_queryset(self) -> QuerySet:
        return ProjectExpense.objects.\
            filter(budget_item__project__project_group__company=self.request.user.company)
