from django.db import models
from django.core.validators import MaxValueValidator, MinValueValidator
from djmoney.models.fields import MoneyField
from core.models import BaseTimestampedModel
from django.db.models import F


class Issue(BaseTimestampedModel):
    class SeverityChoices(models.IntegerChoices):
        MINOR = 1, 'незначительные'
        ESSENTIAL = 2, 'существенные'
        MAJOR = 3, 'высокие'
        CATASTROPHICAL = 4, 'катастрофические'

    class ControlMethodChoices(models.TextChoices):
        ACCEPT = 'ACCEPT', 'принятие'
        DECREASE = 'DECREASE', 'снижение'
        TRANSFER = 'TRANSFER', 'передача'
        DENY = 'DENY', 'отказ'

    company = models.ForeignKey(
        'companies.Company',
        on_delete=models.CASCADE,
        verbose_name='компания',
        related_name='issues',
    )
    project = models.ForeignKey(
        'projects.Project',
        on_delete=models.CASCADE,
        verbose_name='проект',
        related_name='issues',
        blank=True,
        null=True,
    )
    category = models.ForeignKey(
        'IssueCategory',
        on_delete=models.PROTECT,
        verbose_name='категория',
        related_name="issues",
        blank=True,
        null=True
    )
    name = models.CharField('название', max_length=255)
    description = models.TextField('описание', blank=True, null=True)
    probability = models.PositiveIntegerField(
        'Вероятность наступления(в %)',
        validators=[MinValueValidator(0), MaxValueValidator(100)],
        blank=True,
        null=True
    )
    author = models.ForeignKey(
        'users.User',
        on_delete=models.PROTECT,
        verbose_name='автор',
        related_name='created_issues'
    )
    severity = models.PositiveIntegerField(
        'Степень последствий',
        validators=[MinValueValidator(0), MaxValueValidator(5)],
        blank=True,
        null=True,
        choices=SeverityChoices.choices,
    )
    control_method = models.CharField(
        'способ управления',
        blank=True,
        null=True,
        max_length=10,
        choices=ControlMethodChoices.choices,
    )
    consequences = models.TextField('последствия', blank=True, null=True)
    response = models.TextField('сценарий реагирования', blank=True, null=True)
    damage = MoneyField('ущерб', null=True, blank=True, default_currency=None, max_digits=19, decimal_places=4)
    responsible_user = models.ForeignKey(
        'users.User',
        on_delete=models.SET_NULL,
        verbose_name='ответственный',
        blank=True,
        null=True,
        related_name='responsible_for_issues'
    )
    domains = models.ManyToManyField(
        'IssueDomain',
        verbose_name='области проблемы',
        blank=True,
        null=True,
        related_name='issues'
    )

    order = models.PositiveIntegerField(
        'порядок',
        default=10,
        blank=False,
        null=False,
    )
    cause = models.TextField('причина', blank=True, null=True)
    root_cause = models.TextField('первопричина', blank=True, null=True)

    class Meta:
        ordering = [F('probability').desc(nulls_last=True), F('severity').desc(nulls_last=True)]

    def __str__(self):
        return f'{self.project} - {self.name}'

    @property
    def is_problem(self) -> bool:
        if self.probability == 100:
            return True
        return False

    @property
    def risk_assessment(self):
        if self.probability is None or self.severity is None:
            return None
        return ((self.probability / 100) * self.severity) * 5 / 4
        # return self.probability * self.severity


class IssueCategory(models.Model):
    name = models.CharField(
        'имя категории',
        max_length=100,
        unique=True
    )
    order = models.IntegerField(
        'порядок',
        default=10
    )
    company = models.ForeignKey(
        'companies.Company',
        on_delete=models.CASCADE,
        verbose_name='компания',
        related_name='issue_categories',
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = 'Issue categories'
        ordering = ['company', 'order']


class IssueDomain(models.Model):
    name = models.CharField(
        'имя области проблемы',
        max_length=100,
        unique=True
    )
    order = models.IntegerField(
        'порядок',
        default=10
    )
    company = models.ForeignKey(
        'companies.Company',
        on_delete=models.CASCADE,
        verbose_name='компания',
        related_name='issue_domains',
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = 'Issue domains'
        ordering = ['company', 'order']
