from django.db import models
from mptt.models import MPTTModel, TreeForeignKey
from django.db.models.signals import post_save
from datetime import datetime
from django.contrib.postgres.fields import ArrayField
from roles.models import UserRole, Role


class ProjectGroup(MPTTModel):
    name = models.CharField('наименование', max_length=255)
    company = models.ForeignKey(
        'companies.Company',
        verbose_name='компания',
        related_name='project_groups',
        on_delete=models.CASCADE,
    )
    parent_group = TreeForeignKey(
        'self',
        verbose_name='родительский портфель',
        related_name='subgroups',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    owner = models.ForeignKey(
        'users.User',
        on_delete=models.CASCADE,
        verbose_name='автор',
        related_name='project_group_owner',
    )

    class MPTTMeta:
        parent_attr = 'parent_group'

    def __str__(self):
        return '{} - {}'.format(self.company.name, self.name)

    @classmethod
    def post_create(cls, sender, instance, created, *args, **kwargs):
        """
        Method for auto-creating default project settings, triggered by django signals.
        """
        if not created:
            return
        admin_role = Role.objects.get(pk=-2)

        new_role = UserRole.objects.create(
            user=instance.owner,
            role=admin_role,
            project_group=instance,
        )


post_save.connect(ProjectGroup.post_create, sender=ProjectGroup)


class Project(models.Model):
    # TODO: add @closed@ status
    ACT = 'ACT'
    FUT = 'FUT'
    CNC = 'CNC'
    CLS = 'CLS'
    PSP = 'PSP'

    STATUS_CHOICES = (
        (FUT, 'Перспективный'),
        (ACT, 'Открытый'),
        (CNC, 'Отмененный'),
        (CLS, 'Завершенный'),
        (PSP, 'Отложенный'),
    )

    name = models.CharField(
        'название проекта',
        max_length=30,
        unique=True,
        db_index=True
    )
    # eng_name = models.CharField(
    #     'английское название (для ссылок)',
    #     max_length=20,
    #     unique=True
    # )
    description = models.TextField('описание', blank=True)
    start_date = models.DateField('начало работы', blank=True, null=True)
    end_date = models.DateField('конец работы', blank=True, null=True)
    budget = models.FloatField('бюджет работы', blank=True, null=True)
    customer = models.ForeignKey(
        'contacts.Partner',
        on_delete=models.PROTECT,
        verbose_name='заказчик',
        related_name='projects_customer',
        blank=True,
        null=True,
    )
    general_customer = models.ForeignKey(
        'contacts.Partner',
        related_name='projects_general_customer',
        on_delete=models.PROTECT,
        verbose_name='генеральный заказчик',
        blank=True,
        null=True,
    )
    abbr = models.CharField('аббревиатура', max_length=10, unique=True)
    project_type = models.ForeignKey(
        'ProjectType',
        on_delete=models.PROTECT,
        verbose_name='тип проекта',
        related_name='projects',
    )
    project_group = models.ForeignKey(
        'ProjectGroup',
        on_delete=models.CASCADE,
        verbose_name='портфель проектов',
        related_name='projects',
    )
    # lead_dept = models.ForeignKey(
    #     'contacts.Department',
    #     on_delete=models.PROTECT,
    #     verbose_name='ведущее подразделение',
    #     blank=True,
    #     null=True
    # )
    status = models.CharField(
        'статус проекта',
        max_length=20,
        db_index=True,
        choices=STATUS_CHOICES,
        default=STATUS_CHOICES[0][0],
    )
    has_problems = models.BooleanField('проблемный', default=False)
    watchers = models.ManyToManyField('users.User', related_name='projects_watched', blank=True)
    owner = models.ForeignKey(
        'users.User',
        on_delete=models.CASCADE,
        verbose_name='автор',
        related_name='project_owner',
    )

    def __str__(self):
        return self.name

    @classmethod
    def post_create(cls, sender, instance, created, *args, **kwargs):
        """
        Method for auto-creating default project settings, triggered by django signals.
        """
        if not created:
            return
        ProjectPreferences.objects.create(
            project=instance,
        )
        admin_role = Role.objects.get(pk=-3)

        new_role = UserRole.objects.create(
            user=instance.owner,
            role=admin_role,
            project=instance,
        )


post_save.connect(Project.post_create, sender=Project)


class ProjectStage(models.Model):
    project = models.ForeignKey('Project', on_delete=models.CASCADE, related_name='stages')
    stage = models.SmallIntegerField('этап')
    name = models.TextField('полное наименование этапа', blank=True)
    start_date = models.DateField('начало этапа')
    end_date = models.DateField('конец этапа')
    budget = models.FloatField('бюджет этапа', blank=True, null=True)
    # order_number = models.CharField('номер заказа', max_length=7, blank=True)
    # order_details = models.TextField('реквизиты заказа', blank=True)

    def __str__(self):
        return u'Этап {} проекта {}'.format(self.stage, self.project)


class ProjectCategory(models.Model):
    name = models.CharField('имя категории', max_length=100, unique=True)
    eng_name = models.CharField('name', max_length=100, unique=True, null=True)
    order = models.IntegerField('порядок', default=10)
    about = models.TextField('описание категории', blank=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['order']
        verbose_name_plural = 'Project categories'


class ProjectType(models.Model):
    name = models.CharField('имя типа', max_length=30)
    category = models.ForeignKey(
        'ProjectCategory',
        on_delete=models.CASCADE,
        verbose_name='категория',
        related_name='types',
    )
    about = models.TextField('описание типа', blank=True)

    def __str__(self):
        return '{}: {}'.format(str(self.category), self.name)

    class Meta:
        ordering = ['category', 'name']
        unique_together = ('category', 'name')


class ProjectPreferences(models.Model):
    project = models.OneToOneField(
        'Project',
        on_delete=models.CASCADE,
        verbose_name='проект',
        related_name='preferences',
    )
    finance = models.BooleanField('финансы', default=True)
    procurements = models.BooleanField('закупки', default=True)
    contractors = models.BooleanField('контрагенты', default=True)
    assignments = models.BooleanField('поручения', default=True)
    products = models.BooleanField('продукты', default=True)
    reports = models.BooleanField('отчеты', default=True)
    questions = models.BooleanField('вопросы', default=True)
    works = models.BooleanField('работы', default=True)
    # TODO: add doc_cats and doc_types, but with reversed logic, so that they should be excluded
    # doc_cats = models.ManyToManyField('documents.Doc_category', verbose_name='категории документов', blank=True)
    # doc_types = models.ManyToManyField('documents.Doc_type', verbose_name='типы документов', blank=True)

    def __str__(self):
        return 'Настройки для проекта {}'.format(self.project)


class ProjectMeta(models.Model):
    string = 'string'
    text = 'text'
    number = 'number'
    date = 'date'
    time = 'time'
    date_time = 'datetime'

    DATA_TYPE_CHOICES = (
        (string, 'Строка'),
        (text, 'Текст'),
        (number, 'Число'),
        (date, 'Дата'),
        (time, 'Время'),
        (date_time, 'Дата со временем'),
    )

    created = models.DateTimeField('дата и время создания', default=datetime.now)
    owner = models.ForeignKey(
        'users.User',
        on_delete=models.CASCADE,
        verbose_name='автор',
        related_name='projects_meta_owner',
        blank=True,
        null=True,
    )
    name = models.CharField('название', max_length=30)
    data_type = models.CharField(
        'тип данных',
        max_length=20,
        db_index=True,
        choices=DATA_TYPE_CHOICES,
        default=DATA_TYPE_CHOICES[0][0],
    )
    project_categories = models.ManyToManyField(
        'projects.ProjectCategory',
        verbose_name='категории проекта',
        blank=True
    )
    public = models.BooleanField('публичный', default=False)

    def __str__(self):
        return self.name


class ProjectMetaValue(models.Model):
    created = models.DateTimeField('дата и время создания', default=datetime.now)
    updated = models.DateTimeField('дата и время обновления', default=datetime.now)
    project = models.ForeignKey(
        'Project',
        on_delete=models.CASCADE,
        verbose_name='Проект',
        blank=True,
        related_name='project_meta_values')
    meta = models.ForeignKey(
        'ProjectMeta',
        on_delete=models.CASCADE,
        verbose_name='Метаданные',
        blank=True,
        related_name='project_meta_values')
    value = models.TextField('значение метаданных', blank=True)

    def __str__(self):
        return '{} - {} - {}'.format(self.project, self.meta, self.value)


class ProjectView(models.Model):
    created = models.DateTimeField('дата и время создания', default=datetime.now)
    updated = models.DateTimeField('дата и время обновления', default=datetime.now)
    name = models.CharField('название представления', max_length=100)
    owner = models.ForeignKey(
        'users.User',
        on_delete=models.CASCADE,
        verbose_name='пользователь',
        related_name='projects_view_owners',
        blank=True,
        null=True,
    )
    public = models.BooleanField('публичный', default=False)
    current = models.BooleanField('текущий', default=False)
    fields = ArrayField(models.CharField(max_length=255, blank=True))
    order_by = ArrayField(models.CharField(max_length=255, blank=True))
    group_by = ArrayField(models.CharField(max_length=255, blank=True), blank=True)
    filter_by = ArrayField(models.CharField(max_length=255, blank=True), blank=True)

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name

    @classmethod
    def post_create(cls, sender, instance, created, *args, **kwargs):
        """
        Method for auto-update project view value current, triggered by django signals.
        """
        try:
            current_project_view = ProjectView.objects.get(owner=instance.owner, current=True)
        except:
            return
        if instance.current and current_project_view and current_project_view.id != instance.id:
            ProjectView.objects.filter(owner=instance.owner, current=True).update(current=False)


post_save.connect(ProjectView.post_create, sender=ProjectView)
