from core.serializers import ModelSerializer
from .models import (
    Project,
    ProjectGroup,
    ProjectCategory,
    ProjectType,
    ProjectStage,
    ProjectPreferences,
    ProjectMeta,
    ProjectMetaValue,
    ProjectView,
)

from companies.serializers import CompanySerializer
from contacts.serializers import PartnerSerializer
from roles.serializers import UserRoleSerializer
from rest_framework import serializers
from users.serializers import UserSerializer
from finance.serializers import ProjectBudgetTypeSerializer, ProjectBudgetSerializer


class ProjectCategorySerializer(ModelSerializer):

    class Meta:
        model = ProjectCategory
        fields = (
            'id',
            'name',
            'eng_name',
            'order',
            'about',
        )


class ProjectTypeSerializer(ModelSerializer):

    class Meta:
        model = ProjectType
        fields = (
            'id',
            'name',
            'category',
            'about',
        )
        relation_fields = [('category', ProjectCategorySerializer)]


class ProjectStageSerializer(ModelSerializer):

    class Meta:
        model = ProjectStage
        fields = (
            'id',
            'project',
            'stage',
            'name',
            'start_date',
            'end_date',
            'budget',
        )


class ProjectPreferencesSerializer(ModelSerializer):

    class Meta:
        model = ProjectPreferences
        fields = '__all__'


class ProjectMetaSerializer(ModelSerializer):

    class Meta:
        model = ProjectMeta
        fields = '__all__'
        relation_fields = [
            ('owner', UserSerializer),
            ('project_categories', ProjectCategorySerializer),
        ]


class ProjectMetaValueSerializer(ModelSerializer):

    class Meta:
        model = ProjectMetaValue
        fields = '__all__'
        relation_fields = [
            ('meta', ProjectMetaSerializer)
        ]


class ProjectBaseSerializer(ModelSerializer):
    users_roles = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)

    class Meta:
        model = Project
        fields = (
            'id',
            'name',
            'description',
            'start_date',
            'end_date',
            'budget',
            'abbr',
            'customer',
            'general_customer',
            'project_type',
            'project_group',
            'status',
            'has_problems',
            'watchers',
            'owner',
            'stages',
            'preferences',
            'users_roles',
            'project_meta_values',
            'project_budget_types',
            'project_budget',
        )
        relation_fields = (
            ('customer', PartnerSerializer),
            ('general_customer', PartnerSerializer),
            ('owner', UserSerializer),
            ('project_type', ProjectTypeSerializer),
            ('users_roles', UserRoleSerializer),
            ('stages', ProjectStageSerializer),
            ('preferences', ProjectPreferencesSerializer),
            ('watchers', UserSerializer),
            ('project_meta_values', ProjectMetaValueSerializer),
            ('project_budget_types', ProjectBudgetTypeSerializer),
            ('project_budget', ProjectBudgetSerializer),
        )


class ProjectGroupBaseSerializer(ModelSerializer):
    projects = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)
    subgroups = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)

    class Meta:
        model = ProjectGroup
        fields = (
            'id',
            'name',
            'company',
            'parent_group',
            'projects',
            'subgroups',
            'owner'
        )


class ProjectGroupSerializer(ModelSerializer):
    projects = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)
    subgroups = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)

    class Meta:
        model = ProjectGroup
        fields = ProjectGroupBaseSerializer.Meta.fields
        relation_fields = [
            ('company', CompanySerializer),
            ('projects', ProjectBaseSerializer),
            ('subgroups', ProjectGroupBaseSerializer),
            ('parent_group', ProjectGroupBaseSerializer)
        ]


class ProjectViewSerializer(ModelSerializer):

    class Meta:
        model = ProjectView
        fields = '__all__'
        relation_fields = [
            ('owner', UserSerializer)
        ]


class ProjectSerializer(ModelSerializer):
    stages = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)
    preferences = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=False)
    users_roles = serializers.PrimaryKeyRelatedField(read_only=True, required=False, many=True)
    project_meta_values = ProjectMetaValueSerializer(read_only=True, required=False, many=True)

    class Meta:
        model = Project
        fields = ProjectBaseSerializer.Meta.fields
        relation_fields = ProjectBaseSerializer.Meta.relation_fields + (
            ('project_group', ProjectGroupSerializer),
            ('project_type', ProjectTypeSerializer),
            ('project_meta_values', ProjectMetaValueSerializer)
        )

