from django.db import models
from django.contrib.auth.models import Permission
from django.core.exceptions import ValidationError
from django.contrib.postgres.fields import ArrayField


class Role(models.Model):
    name = models.CharField('наименование', max_length=255)
    permissions = models.ManyToManyField(
        Permission,
        verbose_name='разрешения',
        blank=True,
        related_name='roles'
    )
    all_permissions = models.BooleanField('все разрешения', default=False)
    company = models.BooleanField('для компании', default=False)
    project_group = models.BooleanField('для портфеля проектов', default=False)
    project = models.BooleanField('для проекта', default=False)
    glob = models.BooleanField('глобальная', default=False)
    widgets = ArrayField(base_field=models.CharField(max_length=255), default=list, null=True, blank=True)

    def __str__(self):
        return self.name

    def clean(self):
        if not any([self.company, self.project, self.project_group, self.glob]):
            raise ValidationError('Нужно выбрать хотя бы одну принадлежность роли')

    def save(self, *args, **kwargs):
        self.full_clean()
        super(Role, self).save(*args, **kwargs)


class UserRole(models.Model):
    user = models.ForeignKey(
        'users.User',
        related_name='roles',
        verbose_name='пользователь',
        on_delete=models.CASCADE,
    )
    role = models.ForeignKey(
        'Role',
        related_name='users_roles',
        verbose_name='роль',
        on_delete=models.CASCADE,
    )
    company = models.ForeignKey(
        'companies.Company',
        related_name='users_roles',
        verbose_name='компания',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project = models.ForeignKey(
        'projects.Project',
        related_name='users_roles',
        verbose_name='проект',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project_group = models.ForeignKey(
        'projects.ProjectGroup',
        related_name='users_roles',
        verbose_name='портфель проектов',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    custom_name = models.CharField('особое название', max_length=255, blank=True)

    def get_role_name(self):
        """
        :return: role name or custom role name, if present
        :rtype: str
        """
        if self.custom_name:
            return self.custom_name
        else:
            return self.role.name

    @property
    def role_name(self):
        return self.get_role_name()

    def get_permission_object(self):
        """
        :return: object, upon which the role is functional, or None
        :rtype: object
        """
        if self.company:
            return self.company
        if self.project_group:
            return self.project_group
        if self.project:
            return self.project
        return None

    def __str__(self):
        return '{} - {}'.format(self.get_role_name(), str(self.get_permission_object()))

