from typing import List

from django.db import models
from django.contrib.auth.models import AbstractUser, Permission, UserManager, AnonymousUser as DefaultAnonymousUser
from projects.models import Project, ProjectGroup
from roles.models import Role
from news.models import NewsEntry

from django.utils.translation import gettext_lazy as _
from core.helpers import get_file_path


# TODO: add Vacation class

class UserManager(UserManager):
    def _create_user(self, email, password, **extra_fields):
        """
        Create and save a user with the given username, email, and password.
        """
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email=None, password=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, password, **extra_fields)


class User(AbstractUser):
    username = models.CharField(
        'имя пользователя',
        max_length=150,
        blank=True,
    )
    email = models.EmailField('email', unique=True)
    middle_name = models.CharField(
        'отчество',
        max_length=255,
        blank=True
    )
    place = models.CharField(
        'кабинет',
        max_length=255,
        blank=True
    )
    deputy = models.ForeignKey(
        'User',
        on_delete=models.SET_NULL,
        verbose_name='заместитель',
        related_name='users_substituted',
        blank=True,
        null=True
    )
    birthday = models.DateField(
        'день рождения',
        blank=True,
        null=True
    )
    company = models.ForeignKey(
        'companies.Company',
        on_delete=models.CASCADE,
        verbose_name='компания',
        related_name='employees',
        blank=True,
        null=True,
    )
    department = models.ForeignKey(
        'companies.Department',
        on_delete=models.SET_NULL,
        verbose_name='департамент',
        related_name='employees',
        blank=True,
        null=True,
    )
    department_manager = models.BooleanField('управляет департаментом', default=False)
    position = models.CharField(
        'должность',
        max_length=255,
        blank=True,
    )
    avatar = models.ImageField(
        _("Avatar"),
        upload_to=get_file_path,
        default='ava.png',
        null=True,
        blank=True
    )

    # TODO: add photo field

    objects = UserManager()

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = ['first_name', 'last_name']

    class Meta:
        ordering = ['last_name', 'first_name', 'middle_name']
        verbose_name = 'пользователь'
        verbose_name_plural = 'пользователи'

    def __str__(self):
        return self.email

    @property
    def full_name(self):
        return self.get_full_name()

    def get_full_name(self):
        """
        Returns full user name: first name, middle name, last name
        i.e.: Александр Борисович Иванов

        :return: full user name
        :rtype: str
        """
        if self.middle_name:
            return '{} {} {}'.format(self.first_name, self.middle_name, self.last_name)
        else:
            return '{} {}'.format(self.first_name, self.last_name)

    @property
    def short_name(self):
        return self.get_short_name()

    def get_short_name(self):
        """
        Returns shortened user name: first letters of first and middle names and full last name.
        i.e.: А.Б. Иванов

        :return: Shortened user name
        :rtype: str
        """
        if self.middle_name:
            return '{:.1}.{:.1}. {}'.format(self.first_name, self.middle_name, self.last_name)
        else:
            return '{:.1}. {}'.format(self.first_name, self.last_name)

    def dismiss(self):
        """
        User dismiss: set is_active to false, forbids user login,
        change username to format: "dismissed_user_<user_id>_<old_username>"

        :return: new username
        :rtype: str
        """
        self.is_active = False
        self.set_unusable_password()
        self.email = 'dismissed_user_' + str(self.id) + '_' + self.email
        self.save()
        return self.email

    def has_permission(self, permission, company=None, project_group=None, project=None):
        """
        Checks if user has indicated permission.
        Default search is in global permissions.
        If company is specified, checks for permission in the realm of the indicated
        company, otherwise uses user's company field with regards to company hierarchy.
        If project_group is specified, checks for permission in the realm of indicated
        project_group with regards to project_group hierarchy and with regards to
        project_group company and its own permissions and hierarchy
        If project is specified, checks for permission in the realm of indicated
        project, with regards to its project_group and its hierarchy.

        NOTE: Only one of the three is meant to be provided, as any combinations
        do not make sense.
        If more than one optional parameter is given, only lower-tier parameter will be
        taken into account.
        If user is inactive or does not have usable password, this will always return False

        :param permission: permission codename
        :param company: company, can be Company object or id (id not implemented)
        :param project_group: project_group, can be ProjectGroup object or id (id not implemented)
        :param project: project, can be Project object or id (id not implemented)
        :return: True if permission check was successful, False otherwise.
        """
        if not self.is_active or not self.has_usable_password():
            return False
        if self.is_superuser:
            return True

        # Global roles
        roles = self.roles.filter(role__glob=True)

        # Company roles
        if project:
            companies = project.project_group.company.get_ancestors(include_self=True)
        elif project_group:
            companies = project_group.company.get_ancestors(include_self=True)
        elif company:
            companies = company.get_ancestors(include_self=True)
        elif self.company:
            companies = self.company.get_ancestors(include_self=True)
        else:
            companies = None
        if companies:
            roles = roles | self.roles.filter(role__company=True, company__in=companies)

        # ProjectGroup roles
        if project:
            pgs = project.project_group.get_ancestors(include_self=True)
        elif project_group:
            pgs = project_group.get_ancestors(include_self=True)
        else:
            pgs = None

        if pgs:
            roles = roles | self.roles.filter(role__project_group=True, project_group__in=pgs)

        # Project roles
        if project:
            roles = roles | self.roles.filter(role__project=True, project=project)

        # Permission check
        if roles.filter(role__all_permissions=True):
            return True
        if roles.filter(role__permissions__codename=permission):
            return True
        return False

    def get_projects(self, status=None):
        """
        Gets all projects in which the user participates.

        :param status: only return projects with indicated status, ACT, FUT or ARCH
        :return: Project queryset
        """
        if status:
            return Project.objects.filter(users_roles__user=self, status=status).distinct()
        else:
            return Project.objects.filter(users_roles__user=self).distinct()

    def get_projects_user_watcher(self, status=None):
        """
        Gets all projects in which the user is an watcher.

        :param status: only return projects with indicated status, ACT, FUT or ARCH
        :return: Project queryset
        """
        if status:
            return Project.objects.filter(watchers=self.id, status=status).distinct()
        else:
            return Project.objects.filter(watchers=self.id).distinct()

    def get_project_groups(self):
        """
        Gets all project groups in which the user participates

        :return ProjectGroup queryset
        """
        return ProjectGroup.objects.filter(users_roles__user=self).distinct()

    def get_projects_news(self):
        return NewsEntry.objects.filter(project__in=self.get_projects())

    def get_department_news(self):
        return NewsEntry.objects.filter(departament=self.department)

    def get_company_news(self):
        return NewsEntry.objects.filter(company=self.company)

    # TODO: write unit test for this method
    def get_user_permissions(self, company=None, project_group=None, project=None):
        """
        Gets user permissions in the realm defined by parameters.

        If company is specified, gets permissions in the realm of the indicated
        company, otherwise uses user's company field with regards to company hierarchy.
        If project_group is specified, gets permissions in the realm of indicated
        project_group with regards to project_group hierarchy and with regards to
        project_group company and its own permissions and hierarchy
        If project is specified, gets permissions in the realm of indicated
        project, with regards to its project_group and its hierarchy.

        NOTE: Only one of the three is meant to be provided, as any combinations
        do not make sense.
        If more than one optional parameter is given, only lower-tier parameter will be
        taken into account.
        If user is inactive or does not have usable password, this will always return empty list

        :param company: company
        :param project_group: project group
        :param project: project
        :return: list of permissions codenames
        """
        if not self.is_active or not self.has_usable_password():
            return []
        if self.is_superuser:
            return Permission.objects.all().values_list('codename', flat=True)

        all_user_roles = Role.objects.filter(users_roles__user=self)

        # Global roles
        roles = all_user_roles.filter(glob=True)
        # roles = self.roles.filter(role__glob=True)

        # Company roles
        if project:
            companies = project.project_group.company.get_ancestors(include_self=True)
        elif project_group:
            companies = project_group.company.get_ancestors(include_self=True)
        elif company:
            companies = company.get_ancestors(include_self=True)
        elif self.company:
            companies = self.company.get_ancestors(include_self=True)
        else:
            companies = None
        if companies:
            # roles = roles | self.roles.filter(role__company=True, company__in=companies)
            roles = roles | all_user_roles.filter(company=True, users_roles__company__in=companies)

        # ProjectGroup roles
        if project:
            pgs = project.project_group.get_ancestors(include_self=True)
        elif project_group:
            pgs = project_group.get_ancestors(include_self=True)
        else:
            pgs = None

        if pgs:
            # roles = roles | self.roles.filter(role__project_group=True, project_group__in=pgs)
            roles = roles | all_user_roles.filter(project_group=True, users_roles__project_group__in=pgs)


        # Project roles
        if project:
            # roles = roles | self.roles.filter(role__project=True, project=project)
            roles = roles | all_user_roles.filter(project=True, users_roles__project=project)

        if roles.filter(all_permissions=True):
            return Permission.objects.all().values_list('codename', flat=True)

        permissions = Permission.objects.filter(roles__in=roles)

        return permissions.distinct().values_list('codename', flat=True)


class UserLastProjectView(models.Model):
    user = models.ForeignKey(
        'User',
        on_delete=models.CASCADE,
        verbose_name='пользователь',
        related_name='last_project_views',
        blank=True,
        null=True,
    )
    view = models.ForeignKey(
        'projects.ProjectView',
        on_delete=models.CASCADE,
        verbose_name='представление',
        related_name='last_project_views',
        blank=True,
        null=True,
    )

    def __str__(self):
        return '{} - {}'.format(self.user, self.view)


class AnonymousUser(DefaultAnonymousUser):
    company = None
