from typing import Dict

from django.contrib.auth.models import Permission
from django.test import TestCase

from companies.models import Company
from roles.models import Role, UserRole
from .models import User
from projects.models import Project, ProjectGroup, ProjectCategory, ProjectType


class UsersTestCase(TestCase):
    def setUp(self):
        User.objects.create(
            email='ivanov',
            first_name='Иван',
            middle_name='Валентинович',
            last_name='Иванов',
        )
        User.objects.create(
            email='petrov',
            first_name='Сергей',
            last_name='Петров',
        )

    def test_user_full_name(self):
        ivanov = User.objects.get(email='ivanov')
        petrov = User.objects.get(email='petrov')
        self.assertEqual(ivanov.get_full_name(), 'Иван Валентинович Иванов')
        self.assertEqual(petrov.get_full_name(), 'Сергей Петров')

    def test_user_short_name(self):
        ivanov = User.objects.get(email='ivanov')
        petrov = User.objects.get(email='petrov')
        self.assertEqual(ivanov.get_short_name(), 'И.В. Иванов')
        self.assertEqual(petrov.get_short_name(), 'С. Петров')

    def test_user_dismissal(self):
        ivanov = User.objects.get(email='ivanov')
        new_email = ivanov.dismiss()
        self.assertEqual(new_email, 'dismissed_user_' + str(ivanov.id) + '_ivanov')

        d_ivanov = User.objects.get(email=new_email)
        self.assertEqual(d_ivanov.is_active, False)
        self.assertEqual(d_ivanov.has_usable_password(), False)


class UserHasPermissionsTestCase(TestCase):
    def setUp(self):
        self.companies = {
            'head':        Company.objects.create(name='head', full_name='head'),
            'subhead':     Company.objects.create(name='subhead', full_name='subhead'),
            'subsubhead1': Company.objects.create(name='subsubhead1', full_name='subsubhead1'),
            'subsubhead2': Company.objects.create(name='subsubhead2', full_name='subsubhead2'),
        }
        self.companies['subhead'].parent_company = self.companies['head']
        self.companies['subhead'].save()
        self.companies['subsubhead1'].parent_company = self.companies['subhead']
        self.companies['subsubhead1'].save()
        self.companies['subsubhead2'].parent_company = self.companies['subhead']
        self.companies['subsubhead2'].save()

        self.permissions = {
            'change_project': Permission.objects.get(codename='change_project'),
            'change_logentry': Permission.objects.get(codename='change_logentry'),
            '__all__': Permission.objects.get(codename='__all__'),
        }

        self.users = {
            'superuser': User.objects.create(email='aadmin', is_superuser=True),
            'global_perm': User.objects.create(email='global_perm'),
            'inactive_user': User.objects.create(email='inactive', is_active=False),

            'comp_head': User.objects.create(email='comp_head', company=self.companies['head']),
            'comp_subhead': User.objects.create(email='comp_subhead', company=self.companies['subhead']),
            'comp_subsubhead1': User.objects.create(email='comp_subsubhead1', company=self.companies['subsubhead1']),
            'comp_subsubhead2': User.objects.create(email='comp_subsubhead2'),
            'comp_all': User.objects.create(email='comp_all', company=self.companies['head']),

            'pg_1': User.objects.create(email='pg_1', company=self.companies['head']),
            'pg_2': User.objects.create(email='pg_2', company=self.companies['head']),
            'pg_3': User.objects.create(email='pg_3', company=self.companies['subhead']),
            'pg_all': User.objects.create(email='pg_all', company=self.companies['head']),

            'pr_1': User.objects.create(email='pr_1', company=self.companies['head']),
            'pr_2': User.objects.create(email='pr_2', company=self.companies['head']),
            'pr_all': User.objects.create(email='pr_all', company=self.companies['head']),
        }

        self.project_groups = {
            'head_pg': ProjectGroup.objects.create(name='head_pg', company=self.companies['head']),
            'head_subpg': ProjectGroup.objects.create(name='head_subpg', company=self.companies['head']),
            'subhead_pg': ProjectGroup.objects.create(name='subhead_pg', company=self.companies['subhead']),
        }
        self.project_groups['head_subpg'].parent_group = self.project_groups['head_pg']
        self.project_groups['head_subpg'].save()

        pr_tp = ProjectType.objects.create(
            name='Test project type',
            category=ProjectCategory.objects.create(
                name='Test project category',
                eng_name='TPC',
            )
        )
        self.projects = {
            'project1': Project.objects.create(
                name='project1',
                abbr='pr1',
                project_type=pr_tp,
                project_group=self.project_groups['head_pg'],
            ),
            'project2': Project.objects.create(
                name='project2',
                abbr='pr2',
                project_type=pr_tp,
                project_group=self.project_groups['head_subpg'],
            )
        }

        self.roles = {
            'glob': Role.objects.create(name='global', glob=True),

            'comp1': Role.objects.create(name='comp1', company=True),
            'comp2': Role.objects.create(name='comp2', company=True),
            'comp3': Role.objects.create(name='comp3', company=True),
            'comp_all': Role.objects.create(name='comp_all', company=True, all_permissions=True),


            'pg_1': Role.objects.create(name='pg_1', project_group=True),
            'pg_2': Role.objects.create(name='pg_2', project_group=True),
            'pg_all': Role.objects.create(name='pg_all', project_group=True, all_permissions=True),

            'pr_1': Role.objects.create(name='pr_1', project=True),
            'pr_2': Role.objects.create(name='pr_2', project=True),
            'pr_all': Role.objects.create(name='pr_all', project=True, all_permissions=True),
        }
        self.roles['glob'].permissions.add(self.permissions['change_project'])
        self.roles['comp1'].permissions.add(self.permissions['change_project'])
        self.roles['comp2'].permissions.add(self.permissions['change_logentry'])
        self.roles['comp3'].permissions.add(self.permissions['change_logentry'], self.permissions['change_project'])
        self.roles['pg_1'].permissions.add(self.permissions['change_logentry'], self.permissions['change_project'])
        self.roles['pg_2'].permissions.add(self.permissions['change_project'])
        self.roles['pr_1'].permissions.add(self.permissions['change_logentry'], self.permissions['change_project'])
        self.roles['pr_2'].permissions.add(self.permissions['change_project'])

        # glob
        UserRole.objects.create(
            user=self.users['global_perm'],
            role=self.roles['glob'],
        )

        # company
        UserRole.objects.create(
            user=self.users['comp_head'],
            role=self.roles['comp1'],
            company=self.companies['head'],
        )
        UserRole.objects.create(
            user=self.users['comp_head'],
            role=self.roles['comp2'],
            company=self.companies['subsubhead2'],
        )
        UserRole.objects.create(
            user=self.users['comp_subhead'],
            role=self.roles['comp2'],
            company=self.companies['subhead'],
        )
        UserRole.objects.create(
            user=self.users['comp_subsubhead1'],
            role=self.roles['comp3'],
            company=self.companies['subsubhead1'],
        )
        UserRole.objects.create(
            user=self.users['comp_subsubhead2'],
            role=self.roles['comp2'],
            company=self.companies['subsubhead2'],
        )
        UserRole.objects.create(
            user=self.users['comp_all'],
            role=self.roles['comp_all'],
            company=self.companies['subhead'],
        )

        # project_group
        UserRole.objects.create(
            user=self.users['pg_1'],
            role=self.roles['pg_1'],
            project_group=self.project_groups['head_pg'],
        )
        UserRole.objects.create(
            user=self.users['pg_2'],
            role=self.roles['pg_2'],
            project_group=self.project_groups['head_pg'],
        )
        UserRole.objects.create(
            user=self.users['pg_3'],
            role=self.roles['pg_1'],
            project_group=self.project_groups['head_subpg'],
        )
        UserRole.objects.create(
            user=self.users['pg_1'],
            role=self.roles['pg_2'],
            project_group=self.project_groups['subhead_pg'],
        )
        UserRole.objects.create(
            user=self.users['pg_all'],
            role=self.roles['pg_all'],
            project_group=self.project_groups['head_pg'],
        )


        # project
        UserRole.objects.create(
            user=self.users['pr_1'],
            role=self.roles['pr_1'],
            project=self.projects['project1'],
        )
        UserRole.objects.create(
            user=self.users['pr_1'],
            role=self.roles['pr_2'],
            project=self.projects['project2'],
        )
        UserRole.objects.create(
            user=self.users['pr_2'],
            role=self.roles['pr_2'],
            project=self.projects['project1'],
        )
        UserRole.objects.create(
            user=self.users['pr_all'],
            role=self.roles['pr_all'],
            project=self.projects['project1'],
        )

    def test_global_permissions(self):
        self.assertEqual(self.users['superuser'].has_permission('change_project'), True)
        self.assertEqual(self.users['inactive_user'].has_permission('change_project'), False)
        self.assertEqual(self.users['global_perm'].has_permission('change_project'), True)
        self.assertEqual(self.users['global_perm'].has_permission('change_logentry'), False)

    def test_company_permissions(self):
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_project',
            ),
            True
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_project',
                company=self.companies['subsubhead2']),
            True
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_logentry',
                company=self.companies['head']),
            False
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_logentry',
                company=self.companies['subhead']),
            False
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead2']),
            True
        )

        self.assertEqual(
            self.users['comp_subhead'].has_permission(
                'change_logentry',
            ),
            True
        )
        self.assertEqual(
            self.users['comp_subhead'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead1']),
            True
        )
        self.assertEqual(
            self.users['comp_subhead'].has_permission(
                'change_project',
                company=self.companies['subsubhead2']),
            False
        )
        self.assertEqual(
            self.users['comp_subhead'].has_permission(
                'change_logentry',
                company=self.companies['head']),
            False
        )

        self.assertEqual(
            self.users['comp_subsubhead1'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead1']),
            True
        )
        self.assertEqual(
            self.users['comp_subsubhead1'].has_permission(
                'change_project',
            ),
            True
        )
        self.assertEqual(
            self.users['comp_subsubhead1'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead2']),
            False
        )
        self.assertEqual(
            self.users['comp_subsubhead1'].has_permission(
                'change_logentry',
                company=self.companies['subhead']),
            False
        )
        self.assertEqual(
            self.users['comp_subsubhead1'].has_permission(
                'change_logentry',
                company=self.companies['head']),
            False
        )

        self.assertEqual(
            self.users['comp_subsubhead2'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead2']),
            True
        )
        self.assertEqual(
            self.users['comp_subsubhead2'].has_permission(
                'change_logentry',
            ),
            False
        )
        self.assertEqual(
            self.users['comp_subsubhead2'].has_permission(
                'change_logentry',
                company=self.companies['subsubhead1']),
            False
        )
        self.assertEqual(
            self.users['global_perm'].has_permission(
                'change_project',
                company=self.companies['subsubhead1']),
            True
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_project',
                company=self.companies['subhead']),
            True
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_project',
                company=self.companies['subsubhead1']),
            True
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_project',
                company=self.companies['head']),
            False
        )

    def test_project_group_permissions(self):
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_project',
                project_group=self.project_groups['head_pg']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_pg']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_subpg']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_logentry',
                project_group=self.project_groups['subhead_pg']
            ),
            False
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_project',
                project_group=self.project_groups['subhead_pg']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_2'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_pg']
            ),
            False
        )
        self.assertEqual(
            self.users['pg_2'].has_permission(
                'change_logentry',
                project_group=self.project_groups['subhead_pg']
            ),
            False
        )
        self.assertEqual(
            self.users['pg_3'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_pg']
            ),
            False
        )
        self.assertEqual(
            self.users['pg_3'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_subpg']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_3'].has_permission(
                'change_logentry',
                project_group=self.project_groups['subhead_pg']
            ),
            False
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_logentry',
                project_group=self.project_groups['head_pg']),
            False
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_project',
                project_group=self.project_groups['head_pg']),
            True
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_project',
                project_group=self.project_groups['head_subpg']),
            True
        )
        self.assertEqual(
            self.users['pg_all'].has_permission(
                'change_project',
                project_group=self.project_groups['head_pg']),
            True
        )
        self.assertEqual(
            self.users['pg_all'].has_permission(
                'change_project',
                project_group=self.project_groups['head_subpg']),
            True
        )
        self.assertEqual(
            self.users['pg_all'].has_permission(
                'change_project',
                project_group=self.project_groups['subhead_pg']),
            False
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_project',
                project_group=self.project_groups['head_subpg']),
            False
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_project',
                project_group=self.project_groups['subhead_pg']),
            True
        )

    def test_project_permissions(self):
        self.assertEqual(
            self.users['pr_1'].has_permission(
                'change_project',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pr_1'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pr_1'].has_permission(
                'change_logentry',
                project=self.projects['project2']
            ),
            False
        )
        self.assertEqual(
            self.users['pr_2'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            False
        )
        self.assertEqual(
            self.users['pr_2'].has_permission(
                'change_project',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_1'].has_permission(
                'change_logentry',
                project=self.projects['project2']
            ),
            True
        )
        self.assertEqual(
            self.users['comp_head'].has_permission(
                'change_project',
                project=self.projects['project2']
            ),
            True
        )
        self.assertEqual(
            self.users['pr_all'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pr_all'].has_permission(
                'change_logentry',
                project=self.projects['project2']
            ),
            False
        )
        self.assertEqual(
            self.users['pg_all'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            True
        )
        self.assertEqual(
            self.users['pg_all'].has_permission(
                'change_logentry',
                project=self.projects['project2']
            ),
            True
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_logentry',
                project=self.projects['project1']
            ),
            False
        )
        self.assertEqual(
            self.users['comp_all'].has_permission(
                'change_logentry',
                project=self.projects['project2']
            ),
            False
        )


class UserProjectsTestCase(TestCase):
    def setUp(self):
        self.users = {
            'ivanov': User.objects.create(email='ivanov'),
            'petrov': User.objects.create(email='petrov'),
            'sidorov': User.objects.create(email='sidorov'),
        }
        self.company = Company.objects.create(name='head', full_name='head')
        self.project_group = ProjectGroup.objects.create(name='head_pg', company=self.company)

        pr_tp = ProjectType.objects.create(
            name='Test project type',
            category=ProjectCategory.objects.create(
                name='Test project category',
                eng_name='TPC',
            )
        )
        self.projects = {
            'project1': Project.objects.create(
                name='project1',
                abbr='pr1',
                project_type=pr_tp,
                project_group=self.project_group,
            ),
            'project2': Project.objects.create(
                name='project2',
                abbr='pr2',
                project_type=pr_tp,
                project_group=self.project_group,
                status='FUT',
            ),
            'project3': Project.objects.create(
                name='project3',
                abbr='pr3',
                project_type=pr_tp,
                project_group=self.project_group,
                status='ARCH',
            )
        }
        self.roles = {
            'pr_1': Role.objects.create(name='pr_1', project=True),
            'pr_2': Role.objects.create(name='pr_2', project=True),
        }
        UserRole.objects.create(
            user=self.users['ivanov'],
            role=self.roles['pr_1'],
            project=self.projects['project1'],
        )
        UserRole.objects.create(
            user=self.users['ivanov'],
            role=self.roles['pr_2'],
            project=self.projects['project1'],
        )
        UserRole.objects.create(
            user=self.users['petrov'],
            role=self.roles['pr_1'],
            project=self.projects['project1'],
        )
        UserRole.objects.create(
            user=self.users['petrov'],
            role=self.roles['pr_1'],
            project=self.projects['project2'],
        )
        UserRole.objects.create(
            user=self.users['petrov'],
            role=self.roles['pr_2'],
            project=self.projects['project3'],
        )

    def test_get_projects(self):
        self.assertQuerysetEqual(
            self.users['ivanov'].get_projects(),
            [
                self.projects['project1'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrov'].get_projects(),
            [
                self.projects['project1'],
                self.projects['project2'],
                self.projects['project3'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrov'].get_projects(status='FUT'),
            [
                self.projects['project1'],
                self.projects['project2'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrov'].get_projects(status='INVALID_PROJECT_STATUS'),
            [],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrov'].get_projects(status='ARCH'),
            [
                self.projects['project3'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['sidorov'].get_projects(),
            [],
            ordered=False,
            transform=lambda x: x,
        )


class UserProjectGroupsTestCase(TestCase):
    def setUp(self):
        self.users: Dict[str, User] = {
            'ivanov': User.objects.create(email='ivanov'),
            'petrov': User.objects.create(email='petrov'),
            'sidorov': User.objects.create(email='sidorov'),
            'ivanova': User.objects.create(email='ivanova'),
            'petrova': User.objects.create(email='petrova'),
            'sidorova': User.objects.create(email='sidorova'),
        }
        self.companies: Dict[str, Company] = {
            'head': Company.objects.create(name='head', full_name='head'),
            'subhead': Company.objects.create(name='subhead', full_name='subhead'),
            'head2': Company.objects.create(name='head2', full_name='head2'),
        }
        self.companies['subhead'].parent_company = self.companies['head']
        self.companies['subhead'].save()

        self.project_groups: Dict[str, ProjectGroup] = {
            'head_pg': ProjectGroup.objects.create(name='head_pg', company=self.companies['head']),
            'head_subpg': ProjectGroup.objects.create(name='head_subpg', company=self.companies['head']),
            'head_subsubpg1': ProjectGroup.objects.create(name='head_subsubpg1', company=self.companies['head']),
            'head_subsubpg2': ProjectGroup.objects.create(name='head_subsubpg2', company=self.companies['head']),
            'head2_pg': ProjectGroup.objects.create(name='head2_pg', company=self.companies['head2']),
            'subhead_pg': ProjectGroup.objects.create(name='subhead_pg', company=self.companies['subhead']),
            'subhead_subpg': ProjectGroup.objects.create(name='subhead_subpg', company=self.companies['subhead']),
        }

        self.project_groups['head_subpg'].parent_group = self.project_groups['head_pg']
        self.project_groups['head_subpg'].save()
        self.project_groups['head_subsubpg1'].parent_group = self.project_groups['head_subpg']
        self.project_groups['head_subsubpg1'].save()
        self.project_groups['head_subsubpg2'].parent_group = self.project_groups['head_subpg']
        self.project_groups['head_subsubpg2'].save()
        self.project_groups['subhead_subpg'].parent_group = self.project_groups['subhead_pg']
        self.project_groups['subhead_subpg'].save()

        self.roles: Dict[str, Role] = {
            'pgr_1': Role.objects.create(name='pgr_1', project_group=True),
            'pgr_2': Role.objects.create(name='pgr_2', project_group=True),
        }

        UserRole.objects.create(
            user=self.users['ivanov'],
            role=self.roles['pgr_1'],
            project_group=self.project_groups['head_pg'],
        )
        UserRole.objects.create(
            user=self.users['petrov'],
            role=self.roles['pgr_1'],
            project_group=self.project_groups['head_subpg'],
        )
        UserRole.objects.create(
            user=self.users['petrov'],
            role=self.roles['pgr_1'],
            project_group=self.project_groups['head_subsubpg1'],
        )
        UserRole.objects.create(
            user=self.users['sidorov'],
            role=self.roles['pgr_2'],
            project_group=self.project_groups['head2_pg'],
        )
        UserRole.objects.create(
            user=self.users['ivanova'],
            role=self.roles['pgr_2'],
            project_group=self.project_groups['subhead_pg'],
        )
        UserRole.objects.create(
            user=self.users['petrova'],
            role=self.roles['pgr_2'],
            project_group=self.project_groups['subhead_subpg'],
        )

    def test_get_projects(self):
        self.assertQuerysetEqual(
            self.users['ivanov'].get_project_groups(),
            [
                self.project_groups['head_pg'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrov'].get_project_groups(),
            [
                self.project_groups['head_subpg'],
                self.project_groups['head_subsubpg1'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['sidorov'].get_project_groups(),
            [
                self.project_groups['head2_pg'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['ivanova'].get_project_groups(),
            [
                self.project_groups['subhead_pg'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['petrova'].get_project_groups(),
            [
                self.project_groups['subhead_subpg'],
            ],
            ordered=False,
            transform=lambda x: x,
        )
        self.assertQuerysetEqual(
            self.users['sidorova'].get_project_groups(),
            [],
            ordered=False,
            transform=lambda x: x,
        )
