from django.db.models import QuerySet
from drf_yasg.openapi import Parameter
from drf_yasg.utils import swagger_auto_schema

from core.mixins import RelationFieldsViewMixin
from .models import User, UserLastProjectView
from companies.models import Company
from projects.models import ProjectGroup, Project, ProjectType
from .serializers_extra import UserNewsSerializer, UserPermissionsSerializer, UserProjectsSerializer, \
    UserProjectGroupsSerializer, UserAllProjectsSerializer
from rest_framework.decorators import action
from django.shortcuts import get_object_or_404
from news.views import NewsPagination
from news.serializers import NewsEntrySerializer
from .serializers import UserSerializer, UserLastProjectViewSerializer
from core.viewsets import ModelViewSet
from djoser.views import UserViewSet as DjoserUserViewSet
import os
from django.conf import settings


from rest_framework import status
from rest_framework.response import Response
from .serializers import UserAvatarSerializer
from rest_framework.parsers import MultiPartParser, FormParser
from rest_framework import viewsets


class UserViewSet(ModelViewSet):
    queryset = User.objects.none()
    serializer_class = UserSerializer
    filterset_fields = ('department', 'company')

    def get_queryset(self) -> QuerySet:
        return User.objects.filter(company=self.request.user.company)

    @swagger_auto_schema(
        manual_parameters=[
            Parameter('company', 'query', description='Company id', required=False, type='number'),
            Parameter('project_group', 'query', description='Project Group id', required=False, type='number'),
            Parameter('project', 'query', description='Project id', required=False, type='number'),
        ],
        operation_description='Gets user permissions',
    )
    @action(
        detail=True,
        serializer_class=UserPermissionsSerializer,
    )
    def permissions(self, request, pk):
        user = self.get_object()

        company = request.query_params.get('company', None)
        if company:
            company = get_object_or_404(Company, id=company)
        project_group = request.query_params.get('project_group', None)
        if project_group:
            project_group = get_object_or_404(ProjectGroup, id=project_group)
        project = request.query_params.get('project', None)
        if project:
            project = get_object_or_404(Project, id=project)

        return Response({'permissions': list(user.get_user_permissions(
            company=company, project_group=project_group, project=project))})

    @swagger_auto_schema(operation_description='Gets projects in which user is participating')
    @action(
        detail=True,
        serializer_class=UserProjectsSerializer,
    )
    def projects(self, request, pk):
        user = self.get_object()

        serializer = UserProjectsSerializer({'projects': user.get_projects()})

        return Response(serializer.data)


    @swagger_auto_schema(operation_description='Gets all projects user is participating and all projects to which the'
                                               ' user has access to portfolios')
    @action(
        detail=True,
        url_path='projects/all',
        serializer_class=UserProjectsSerializer,
    )
    def projects_all(self, request, pk):
        user = self.get_object()

        serializer_projects_all = UserAllProjectsSerializer({
            'projects': user.get_projects(),
            'project_groups': user.get_project_groups(),
            'projects_user_watcher': user.get_projects_user_watcher(),
        })

        list_projects = []

        for project in serializer_projects_all.data['projects']:
            project['is_my'] = True
            list_projects.append(project)

        for project_group in serializer_projects_all.data['project_groups']:
            pg = dict(project_group)
            if pg['rel_projects']:
                del pg['rel_projects']
            if project_group['rel_projects']:
                for project in project_group['rel_projects']:
                    if project['project_type']:
                        project_type = ProjectType.objects.get(id=project['project_type'])
                        project['rel_project_type'] = {
                            'id': project['project_type'],
                            'name': project_type.name,
                            'category': project_type.category.id
                        }
                    project['rel_project_group'] = pg
                    project['is_my'] = False
                    list_projects.append(project)

        for project_user_watcher in serializer_projects_all.data['projects_user_watcher']:
            project_user_watcher['is_my'] = False
            list_projects.append(project_user_watcher)

        return Response({'projects': self.unique_project_array(list_projects)})

    # У пользователя могут быть личные проекты, которые так же есть в его project_group, и здесь убираются их повторы
    @staticmethod
    def unique_project_array(list_project):
        new_list = []
        id_project_list = []
        for project in list_project:
            if project['id'] not in id_project_list:
                id_project_list.append(project['id'])
                new_list.append(project)
                new_list = sorted(new_list, key=lambda k: k['id'])

        return new_list

    @swagger_auto_schema(operation_description='Gets project groups in which user is participating')
    @action(
        detail=True,
        serializer_class=UserProjectGroupsSerializer,
    )
    def project_groups(self, request, pk):
        user = self.get_object()

        serializer = UserProjectGroupsSerializer({'project_groups': user.get_project_groups()})

        return Response(serializer.data)

    @swagger_auto_schema(operation_desqription='Gets news on projects in which user is participating')
    @action(
        detail=False,
        # or False
        serializer_class=UserNewsSerializer,
        # filterset_fields=(),
    )
    def project_news(self, request,):
        # if 'pk' not in self.kwargs or self.kwargs['pk']=='':
        #     self.kwargs['pk'] = request.user.pk

        # user = self.get_object()


        serializer = UserNewsSerializer({'newsEntry':request.user.get_projects_news()},
                                         # many=True,

                                         )

        return Response(serializer.data)


    @swagger_auto_schema(operation_desqription='Gets news on department in which user is participating')
    @action(
        detail=False,
        # or False
        serializer_class=UserNewsSerializer,
        # filterset_fields=(),
    )
    def department_news(self, request, ):
        # if 'pk' not in self.kwargs or self.kwargs['pk']=='':
        #     self.kwargs['pk'] = request.user.pk

        # user = self.get_object()

        serializer = UserNewsSerializer({'newsEntry': request.user.get_department_news()},
                                                # many=True,
                                                )

        return Response(serializer.data)

    @swagger_auto_schema(operation_desqription='Gets news of company in which user is participating')
    @action(
        detail=False,
        # or False
        serializer_class=UserNewsSerializer,
        # filterset_fields=(),
    )
    def company_news(self, request, ):
        # if 'pk' not in self.kwargs or self.kwargs['pk']=='':
        #     self.kwargs['pk'] = request.user.pk

        # user = self.get_object()

        serializer = UserNewsSerializer({'newsEntry': request.user.get_company_news()},
                                        # many=True,
                                        )

        return Response(serializer.data)

    # TODO: тут в сваггере модель без имени, из-за этого в ангуляре оно с некрасивым названием
    #  генерируется (InlineResponse200). Надо разобраться как присвоить ей нормальное название
    @swagger_auto_schema(operation_desqription='Gets all news of user')
    @action(
        detail=False,
        # or False
        serializer_class=NewsEntrySerializer,
        pagination_class=NewsPagination,
        # filterset_fields=(),
    )
    def all_news(self, request, ):
        news = request.user.get_projects_news().union(request.user.get_department_news()).union(request.user.get_company_news())

        paginated_news = self.paginate_queryset(news.order_by("-date"))
        serializer = self.get_serializer(paginated_news,
                                         many=True,
                                         include=['user', 'project', 'company', 'departament'])



        return self.get_paginated_response(serializer.data)
        # serializer = UserNewsSerializer({'newsEntry': paginated_news},
        #                                 # many=True,
        #                                 include=['user', 'project', 'company', 'department']
        #                                 )


        # return Response(serializer.data)


class UserLastProjectViewViewSet(ModelViewSet):
    queryset = UserLastProjectView.objects.all()
    serializer_class = UserLastProjectViewSerializer
    filterset_fields = ('user', 'view')


class UserAvatarViewSet(viewsets.GenericViewSet):
    queryset = User.objects.all()
    serializer_class = UserAvatarSerializer
    filterset_fields = ('department', 'company')
    parser_classes = (FormParser, MultiPartParser)

    @swagger_auto_schema(operation_description='Upload user photo')
    @action(
        methods=['post'],
        detail=True,
        serializer_class=UserAvatarSerializer,
    )
    def update_avatar(self, request, pk):

        user = User.objects.get(id=pk)

        serializer = UserAvatarSerializer(data=request.data, instance=user)
        if serializer.is_valid():
            path = str(settings.MEDIA_ROOT) + '/' + str(user.avatar) if user.avatar is not None else None
            serializer.save()
            if path is not None:
                if len(path.rsplit('/', 2)) > 2 and path.rsplit('/', 2)[2] != 'ava.png':
                    os.remove(path)
            return Response(serializer.data, status=status.HTTP_200_OK)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class MyDjoserUserViewSet(RelationFieldsViewMixin, DjoserUserViewSet):
    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        if self.request.method == 'GET':
            kwargs['include'] = ['company__currencies', 'department']
        return serializer_class(*args, **kwargs)

    @swagger_auto_schema(methods=["get", "put", "patch", "delete"],
                         operation_description='Gets current user',
                         responses={200: UserSerializer(many=False, include='company, department')})
    @action(["get", "put", "patch", "delete"], detail=False)
    def me(self, request, *args, **kwargs):
        return super().me(request, *args, **kwargs)

