from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from users.models import User


class Work(models.Model):
    class WorkType(models.TextChoices):
        TASK = 'TS', _('Task')
        GROUP = 'GR', _('Group')
        MILESTONE = 'MS', _('Milestone')
        ASSIGNMENT = 'AS', _('Assignment')

    class WorkStatus(models.TextChoices):
        CREATED = 'CR', _('Created')
        WIP = 'WP', _('WIP')
        DONE = 'DN', _('Done')
        ACCEPTED = 'AC', _('Accepted')
        CANCELLED = 'CN', _('Cancelled')

    class IndicatorChoices(models.TextChoices):
        CRITICAL = 'CRITICAL', _('CRITICAL')
        WARNING = 'WARNING', _('WARNING')
        NORMAL = 'NORMAL', _('NORMAL')
        GREAT = 'GREAT', _('GREAT')

    name = models.CharField('название работы', max_length=255)
    description = models.TextField('описание работы', blank=True, null=True)
    result = models.TextField('результат работы', blank=True, null=True)
    work_type = models.CharField(
        max_length=2,
        choices=WorkType.choices,
        default=WorkType.TASK,
    )

    status = models.CharField(
        max_length=2,
        choices=WorkStatus.choices,
        default=WorkStatus.CREATED,
    )

    start_base = models.DateField('начало - план')
    start_fc = models.DateField('начало - прогноз', null=True, blank=True)
    start_fact = models.DateField('начало - факт', null=True, blank=True)

    end_base = models.DateField('окончание - план', null=True, blank=True)
    end_fc = models.DateField('окончание - прогноз', null=True, blank=True)
    end_fact = models.DateField('окончание - факт', null=True, blank=True)
    end_val = models.DateField('окончание - валидация', null=True, blank=True)

    created_at = models.DateTimeField('создано в', default=timezone.now)
    created_by = models.ForeignKey('users.User', verbose_name='автор',
                                   on_delete=models.PROTECT, related_name='works_author')

    group = models.ForeignKey('Work', verbose_name='группа', on_delete=models.CASCADE,
                              related_name='works', blank=True, null=True)
    parent = models.ForeignKey('Work', verbose_name='родительская задача', on_delete=models.CASCADE,
                              related_name='subworks', blank=True, null=True)

    project = models.ForeignKey('projects.Project', verbose_name='проект',
                                on_delete=models.CASCADE, related_name='works', null=True, blank=True)
    company = models.ForeignKey('companies.Company', verbose_name='компания',
                                on_delete=models.CASCADE, related_name='works')

    to_user = models.ForeignKey('users.User', verbose_name='исполнитель',
                                on_delete=models.PROTECT, related_name='works_executor', blank=True, null=True)
    val_user = models.ForeignKey('users.User', verbose_name='проверяющий',
                                 on_delete=models.PROTECT, related_name='works_validator', blank=True, null=True)

    done_by = models.ForeignKey('users.User', verbose_name='фактический исполнитель',
                                on_delete=models.PROTECT, related_name='+', blank=True, null=True)
    val_by = models.ForeignKey('users.User', verbose_name='фактический проверяющий',
                               on_delete=models.PROTECT, related_name='+', blank=True, null=True)

    milestone_level = models.PositiveSmallIntegerField('уровень вехи', default=3, blank=True)
    order = models.PositiveSmallIntegerField('порядок', default=0, blank=True)

    disallowed_status_changes = (
        (WorkStatus.ACCEPTED, WorkStatus.CREATED),
        (WorkStatus.ACCEPTED, WorkStatus.CANCELLED),
        (WorkStatus.CANCELLED, WorkStatus.DONE),
        (WorkStatus.DONE, WorkStatus.CREATED),
    )

    def __str__(self):
        return f'{self.id} - {self.project} - {self.name}'

    class Meta:
        ordering = ('order', )
        permissions = [
            ('set_forecast', 'Can set forecast dates'),
            ('split_work', 'Can split works'),
        ]



    @property
    def overdue(self):
        nowdate = timezone.now().date()
        if self.status == self.WorkStatus.CREATED:
            return (self.start_fc or self.start_base) < nowdate
        if self.status == self.WorkStatus.WIP:
            return (self.end_fc or self.end_base) < nowdate
        return False

    def change_status(self, new_status: str, user: User) -> bool:
        if self.status == new_status or (self.status, new_status) in self.disallowed_status_changes:
            return False
        self.status = new_status

        if new_status == self.WorkStatus.CREATED:
            self.done_by = None
            self.val_by = None
            self.end_val = None
            self.end_fact = None
            self.start_fact = None
        if new_status == self.WorkStatus.WIP:
            self.done_by = None
            self.val_by = None
            self.end_val = None
            self.end_fact = None
            self.start_fact = timezone.now().date()
        if new_status == self.WorkStatus.DONE:
            self.done_by = user
            self.val_by = None
            self.end_val = None
            self.end_fact = timezone.now().date()
            if not self.start_fact:
                self.start_fact = timezone.now().date()
        if new_status == self.WorkStatus.ACCEPTED:
            self.val_by = user
            self.end_val = timezone.now().date()
            if not self.end_fact:
                self.end_fact = timezone.now().date()
            if not self.start_fact:
                self.start_fact = timezone.now().date()
            if not self.done_by:
                self.done_by = user

        return True



class WorkLink(models.Model):
    class WorkLinkType(models.TextChoices):
        FTS = '0', _('finish_to_start')
        STS = '1', _('start_to_start')
        FTF = '2', _('finish_to_finish')
        STF = '3', _('start_to_finish')

    source = models.ForeignKey('Work', verbose_name='от', related_name='links_from', on_delete=models.CASCADE)
    target = models.ForeignKey('Work', verbose_name='к', related_name='links_to', on_delete=models.CASCADE)
    link_type = models.CharField(
        max_length=1,
        choices=WorkLinkType.choices,
        default=WorkLinkType.FTS,
    )
    lag = models.IntegerField('сдвиг', blank=True, null=True)



