from typing import Any

from django.db.models import QuerySet
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from core.viewsets import ModelViewSet
from .models import Work, WorkLink
from .serializers import WorkSerializer, WorkLinkSerializer, WorkStatusSerializer
from drf_yasg.utils import swagger_auto_schema
from django_filters import rest_framework as filters


class WorkFilter(filters.FilterSet):
    work_type = filters.BaseInFilter(field_name='work_type')
    status = filters.BaseInFilter(field_name='status')
    project = filters.BaseInFilter(field_name='project')

    class Meta:
        model = Work
        fields = ['work_type', 'status', 'project', 'parent']


class WorksViewSet(ModelViewSet):
    queryset = Work.objects.none()
    serializer_class = WorkSerializer
    filterset_class = WorkFilter

    def get_queryset(self) -> QuerySet:
        return Work.objects.filter(company=self.request.user.company)

    def create(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        request.data['created_by'] = request.user.id
        request.data['company'] = request.user.company.id
        return super(WorksViewSet, self).create(request, *args, **kwargs)

    @swagger_auto_schema(
        operation_description='Change work status',
        responses={200: WorkSerializer(many=False)})
    @action(
        detail=True,
        methods=['POST'],
        serializer_class=WorkStatusSerializer,
    )
    def change_status(self, request, pk):
        work = self.get_object()
        new_status = request.data['status']

        work.change_status(new_status, request.user)
        work.save()

        return Response(WorkSerializer(
                            work,
                            include=['to_user', 'val_user', 'created_by', 'val_by', 'project'],
                            context={'request': request}).data)


class WorkLinksViewSet(ModelViewSet):
    queryset = WorkLink.objects.all()
    serializer_class = WorkLinkSerializer
